# Copyright (C) 2006-2009 Canonical Ltd
# Copyright (C) 2012-2018 Jelmer Vernooij <jelmer@jelmer.uk>

# Authors: Robert Collins <robert.collins@canonical.com>
#          Jelmer Vernooij <jelmer@samba.org>
#          John Carr <john.carr@unrouted.co.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Git-specific subcommands for Bazaar."""

import breezy.bzr  # noqa: F401
from breezy import controldir

from ..commands import Command, display_command
from ..option import Option, RegistryOption


class cmd_git_import(Command):
    """Import all branches from a git repository.

    """

    takes_args = ["src_location", "dest_location?"]

    takes_options = [
        Option('colocated', help='Create colocated branches.'),
        RegistryOption('dest-format',
                       help='Specify a format for this branch. '
                       'See "help formats" for a full list.',
                       lazy_registry=('breezy.controldir', 'format_registry'),
                       converter=lambda name: controldir.format_registry.make_controldir(
                            name),
                       value_switches=True,
                       title="Branch format",
                       ),
        ]

    def _get_colocated_branch(self, target_controldir, name):
        from ..errors import NotBranchError
        try:
            return target_controldir.open_branch(name=name)
        except NotBranchError:
            return target_controldir.create_branch(name=name)

    def _get_nested_branch(self, dest_transport, dest_format, name):
        from ..controldir import ControlDir
        from ..errors import NotBranchError
        head_transport = dest_transport.clone(name)
        try:
            head_controldir = ControlDir.open_from_transport(head_transport)
        except NotBranchError:
            head_controldir = dest_format.initialize_on_transport_ex(
                head_transport, create_prefix=True)[1]
        try:
            return head_controldir.open_branch()
        except NotBranchError:
            return head_controldir.create_branch()

    def run(self, src_location, dest_location=None, colocated=False, dest_format=None):
        import os

        from .. import controldir, trace, ui, urlutils
        from ..controldir import ControlDir
        from ..errors import (BzrError, CommandError, NoRepositoryPresent,
                              NotBranchError)
        from ..i18n import gettext
        from ..repository import InterRepository, Repository
        from ..transport import get_transport
        from .branch import LocalGitBranch
        from .refs import ref_to_branch_name
        from .repository import GitRepository

        if dest_format is None:
            dest_format = controldir.format_registry.make_controldir('default')

        if dest_location is None:
            dest_location = os.path.basename(src_location.rstrip("/\\"))

        dest_transport = get_transport(dest_location)

        source_repo = Repository.open(src_location)
        if not isinstance(source_repo, GitRepository):
            raise CommandError(
                gettext("%r is not a git repository") % src_location)
        try:
            target_controldir = ControlDir.open_from_transport(dest_transport)
        except NotBranchError:
            target_controldir = dest_format.initialize_on_transport_ex(
                dest_transport, shared_repo=True)[1]
        try:
            target_repo = target_controldir.find_repository()
        except NoRepositoryPresent:
            target_repo = target_controldir.create_repository(shared=True)

        if not target_repo.supports_rich_root():
            raise CommandError(
                gettext("Target repository doesn't support rich roots"))

        interrepo = InterRepository.get(source_repo, target_repo)
        mapping = source_repo.get_mapping()
        result = interrepo.fetch()
        with ui.ui_factory.nested_progress_bar() as pb:
            for i, (name, sha) in enumerate(result.refs.items()):
                try:
                    branch_name = ref_to_branch_name(name)
                except ValueError:
                    # Not a branch, ignore
                    continue
                pb.update(gettext("creating branches"), i, len(result.refs))
                if (getattr(target_controldir._format, "colocated_branches",
                            False) and colocated):
                    if name == "HEAD":
                        branch_name = None
                    head_branch = self._get_colocated_branch(
                        target_controldir, branch_name)
                else:
                    head_branch = self._get_nested_branch(
                        dest_transport, dest_format, branch_name)
                revid = mapping.revision_id_foreign_to_bzr(sha)
                source_branch = LocalGitBranch(
                    source_repo.controldir, source_repo, sha)
                if head_branch.last_revision() != revid:
                    head_branch.generate_revision_history(revid)
                source_branch.tags.merge_to(head_branch.tags)
                if not head_branch.get_parent():
                    url = urlutils.join_segment_parameters(
                        source_branch.base,
                        {"branch": urlutils.escape(branch_name)})
                    head_branch.set_parent(url)
        trace.note(gettext(
            "Use 'bzr checkout' to create a working tree in "
            "the newly created branches."))


class cmd_git_object(Command):
    """List or display Git objects by SHA.

    Cat a particular object's Git representation if a SHA is specified.
    List all available SHAs otherwise.
    """

    hidden = True

    aliases = ["git-objects", "git-cat"]
    takes_args = ["sha1?"]
    takes_options = [Option('directory',
                            short_name='d',
                            help='Location of repository.', type=str),
                     Option('pretty', help='Pretty-print objects.')]
    encoding_type = 'exact'

    @display_command
    def run(self, sha1=None, directory=".", pretty=False):
        from ..controldir import ControlDir
        from ..errors import CommandError
        from ..i18n import gettext
        from .object_store import get_object_store
        controldir, _ = ControlDir.open_containing(directory)
        repo = controldir.find_repository()
        object_store = get_object_store(repo)
        with object_store.lock_read():
            if sha1 is not None:
                try:
                    obj = object_store[sha1.encode('ascii')]
                except KeyError:
                    raise CommandError(
                        gettext("Object not found: %s") % sha1)
                if pretty:
                    text = obj.as_pretty_string()
                else:
                    text = obj.as_raw_string()
                self.outf.write(text)
            else:
                for sha1 in object_store:
                    self.outf.write("%s\n" % sha1.decode('ascii'))


class cmd_git_refs(Command):
    """Output all of the virtual refs for a repository.

    """

    hidden = True

    takes_args = ["location?"]

    @display_command
    def run(self, location="."):
        from ..controldir import ControlDir
        from .object_store import get_object_store
        from .refs import get_refs_container
        controldir, _ = ControlDir.open_containing(location)
        repo = controldir.find_repository()
        object_store = get_object_store(repo)
        with object_store.lock_read():
            refs = get_refs_container(controldir, object_store)
            for k, v in sorted(refs.as_dict().items()):
                self.outf.write("%s -> %s\n" %
                                (k.decode('utf-8'), v.decode('utf-8')))


class cmd_git_apply(Command):
    """Apply a series of git-am style patches.

    This command will in the future probably be integrated into "bzr pull".
    """

    takes_options = [
        Option('signoff', short_name='s', help='Add a Signed-off-by line.'),
        Option('force',
               help='Apply patches even if tree has uncommitted changes.')
        ]
    takes_args = ["patches*"]

    def _apply_patch(self, wt, f, signoff):
        """Apply a patch.

        :param wt: A Bazaar working tree object.
        :param f: Patch file to read.
        :param signoff: Add Signed-Off-By flag.
        """
        from dulwich.patch import git_am_patch_split

        from breezy.patch import patch_tree
        (c, diff, version) = git_am_patch_split(f)
        # FIXME: Cope with git-specific bits in patch
        # FIXME: Add new files to working tree
        patch_tree(wt, [diff], strip=1, out=self.outf)
        message = c.message.decode('utf-8')
        if signoff:
            signed_off_by = wt.branch.get_config().username()
            message += "Signed-off-by: {}\n".format(signed_off_by)
        wt.commit(authors=[c.author.decode('utf-8')], message=message)

    def run(self, patches_list=None, signoff=False, force=False):
        from ..errors import UncommittedChanges
        from ..workingtree import WorkingTree
        if patches_list is None:
            patches_list = []

        tree, _ = WorkingTree.open_containing(".")
        if tree.basis_tree().changes_from(tree).has_changed() and not force:
            raise UncommittedChanges(tree)
        with tree.lock_write():
            for patch in patches_list:
                with open(patch) as f:
                    self._apply_patch(tree, f, signoff=signoff)


class cmd_git_push_pristine_tar_deltas(Command):
    """Push pristine tar deltas to a git repository."""

    takes_options = [Option('directory',
                            short_name='d',
                            help='Location of repository.', type=str)]
    takes_args = ['target', 'package']

    def run(self, target, package, directory='.'):
        from ..branch import Branch
        from ..errors import CommandError, NoSuchRevision
        from ..repository import Repository
        from ..trace import warning
        from .mapping import encode_git_path
        from .object_store import get_object_store
        from .pristine_tar import (revision_pristine_tar_data,
                                   store_git_pristine_tar_data)
        source = Branch.open_containing(directory)[0]
        target_bzr = Repository.open(target)
        target = getattr(target_bzr, '_git', None)
        if target is None:
            raise CommandError("Target not a git repository")
        git_store = get_object_store(source.repository)
        with git_store.lock_read():
            tag_dict = source.tags.get_tag_dict()
            for name, revid in tag_dict.iteritems():
                try:
                    rev = source.repository.get_revision(revid)
                except NoSuchRevision:
                    continue
                try:
                    delta, kind = revision_pristine_tar_data(rev)
                except KeyError:
                    continue
                gitid = git_store._lookup_revision_sha1(revid)
                if (not (name.startswith('upstream/') or
                         name.startswith('upstream-'))):
                    warning(
                        "Unexpected pristine tar revision tagged %s. "
                        "Ignoring.", name)
                    continue
                upstream_version = name[len("upstream/"):]
                filename = '{}_{}.orig.tar.{}'.format(
                    package, upstream_version, kind)
                if gitid not in target:
                    warning(
                        "base git id %s for %s missing in target repository",
                        gitid, filename)
                store_git_pristine_tar_data(target, encode_git_path(filename),
                                            delta, gitid)



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.i18n"""

import io

from .. import errors, i18n, tests, workingtree


class ZzzTranslations:
    """Special Zzz translation for debugging i18n stuff.

    This class can be used to confirm that the message is properly translated
    during black box tests.
    """
    _null_translation = i18n._gettext.NullTranslations()

    def zzz(self, s):
        return 'zz\xe5{{%s}}' % s

    def gettext(self, s):
        return self.zzz(self._null_translation.gettext(s))

    def ngettext(self, s, p, n):
        return self.zzz(self._null_translation.ngettext(s, p, n))

    def ugettext(self, s):
        return self.zzz(self._null_translation.ugettext(s))

    def ungettext(self, s, p, n):
        return self.zzz(self._null_translation.ungettext(s, p, n))


class TestZzzTranslation(tests.TestCase):

    def _check_exact(self, expected, source):
        self.assertEqual(expected, source)
        self.assertEqual(type(expected), type(source))

    def test_translation(self):
        trans = ZzzTranslations()

        t = trans.zzz('msg')
        self._check_exact('zz\xe5{{msg}}', t)

        t = trans.gettext('msg')
        self._check_exact('zz\xe5{{msg}}', t)

        t = trans.ngettext('msg1', 'msg2', 0)
        self._check_exact('zz\xe5{{msg2}}', t)
        t = trans.ngettext('msg1', 'msg2', 2)
        self._check_exact('zz\xe5{{msg2}}', t)

        t = trans.ngettext('msg1', 'msg2', 1)
        self._check_exact('zz\xe5{{msg1}}', t)


class TestGetText(tests.TestCase):

    def setUp(self):
        super().setUp()
        self.overrideAttr(i18n, '_translations', ZzzTranslations())

    def test_oneline(self):
        self.assertEqual("zz\xe5{{spam ham eggs}}",
                         i18n.gettext("spam ham eggs"))

    def test_multiline(self):
        self.assertEqual("zz\xe5{{spam\nham\n\neggs\n}}",
                         i18n.gettext("spam\nham\n\neggs\n"))


class TestGetTextPerParagraph(tests.TestCase):

    def setUp(self):
        super().setUp()
        self.overrideAttr(i18n, '_translations', ZzzTranslations())

    def test_oneline(self):
        self.assertEqual("zz\xe5{{spam ham eggs}}",
                         i18n.gettext_per_paragraph("spam ham eggs"))

    def test_multiline(self):
        self.assertEqual("zz\xe5{{spam\nham}}\n\nzz\xe5{{eggs\n}}",
                         i18n.gettext_per_paragraph("spam\nham\n\neggs\n"))


class TestInstall(tests.TestCase):

    def setUp(self):
        super().setUp()
        # Restore a proper env to test translation installation
        self.overrideAttr(i18n, '_translations', None)

    def test_custom_languages(self):
        i18n.install('nl:fy')
        # Whether we found a valid tranlsation or not doesn't matter, we got
        # one and _translations is not None anymore.
        self.assertIsInstance(i18n._translations,
                              i18n._gettext.NullTranslations)

    def test_no_env_variables(self):
        self.overrideEnv('LANGUAGE', None)
        self.overrideEnv('LC_ALL', None)
        self.overrideEnv('LC_MESSAGES', None)
        self.overrideEnv('LANG', None)
        i18n.install()
        # Whether we found a valid tranlsation or not doesn't matter, we got
        # one and _translations is not None anymore.
        self.assertIsInstance(i18n._translations,
                              i18n._gettext.NullTranslations)

    def test_disable_i18n(self):
        i18n.disable_i18n()
        i18n.install()
        # It's disabled, you can't install anything and we fallback to null
        self.assertIsInstance(i18n._translations,
                              i18n._gettext.NullTranslations)


class TestTranslate(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.overrideAttr(i18n, '_translations', ZzzTranslations())

    def test_error_message_translation(self):
        """do errors get translated?"""
        err = None
        tree = self.make_branch_and_tree('.')
        try:
            workingtree.WorkingTree.open('./foo')
        except errors.NotBranchError as e:
            err = str(e)
        self.assertContainsRe(err, "zz\xe5{{Not a branch: .*}}")

    def test_topic_help_translation(self):
        """does topic help get translated?"""
        from .. import help
        out = io.StringIO()
        help.help("authentication", out)
        self.assertContainsRe(
            out.getvalue(), "zz\xe5{{Authentication Settings")


class LoadPluginTranslations(tests.TestCase):

    def test_does_not_exist(self):
        translation = i18n.load_plugin_translations("doesnotexist")
        self.assertEqual("foo", translation.gettext("foo"))



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tag strategies.

These are contained within a branch and normally constructed
when the branch is opened.  Clients should typically do

  Branch.tags.add('name', 'value')
"""

import contextlib
import itertools
import re
import sys
from collections import defaultdict
from typing import Callable, Dict, List, Optional, Set, Tuple

from . import branch as _mod_branch
from . import errors
from .inter import InterObject
from .registry import Registry
from .revision import RevisionID

# NOTE: I was going to call this tags.py, but vim seems to think all files
# called tags* are ctags files... mbp 20070220.




TagSelector = Callable[[str], bool]
TagUpdates = Dict[str, RevisionID]
TagConflict = Tuple[str, bytes, bytes]


def _reconcile_tags(
        source_dict: Dict[str, bytes], dest_dict: Dict[str, bytes],
        overwrite: bool, selector: Optional[TagSelector]) -> Tuple[
            Dict[str, RevisionID], TagUpdates,
            List[TagConflict]]:
    """Do a two-way merge of two tag dictionaries.

    * only in source => source value
    * only in destination => destination value
    * same definitions => that
    * different definitions => if overwrite is False, keep destination
      value and add to conflict list, otherwise use the source value

    :returns: (result_dict, updates,
        [(conflicting_tag, source_target, dest_target)])
    """
    conflicts = []
    updates = {}
    result = dict(dest_dict)  # copy
    for name, target in source_dict.items():
        if selector and not selector(name):
            continue
        if result.get(name) == target:
            pass
        elif name not in result or overwrite:
            updates[name] = target
            result[name] = target
        else:
            conflicts.append((name, target, result[name]))
    return result, updates, conflicts


class Tags:

    def __init__(self, branch):
        self.branch = branch

    def get_tag_dict(self) -> Dict[str, RevisionID]:
        """Return a dictionary mapping tags to revision ids.
        """
        raise NotImplementedError(self.get_tag_dict)

    def get_reverse_tag_dict(self) -> Dict[RevisionID, Set[str]]:
        """Returns a dict with revisions as keys
           and a list of tags for that revision as value"""
        d = self.get_tag_dict()
        rev = defaultdict(set)
        for key in d:
            rev[d[key]].add(key)
        return rev

    def merge_to(self, to_tags: "Tags", overwrite: bool = False,
                 ignore_master: bool = False,
                 selector: Optional[TagSelector] = None) -> Tuple[TagUpdates, Set[TagConflict]]:
        """Copy tags between repositories if necessary and possible.

        This method has common command-line behaviour about handling
        error cases.

        All new definitions are copied across, except that tags that already
        exist keep their existing definitions.

        :param to_tags: Branch to receive these tags
        :param overwrite: Overwrite conflicting tags in the target branch
        :param ignore_master: Do not modify the tags in the target's master
            branch (if any).  Default is false (so the master will be updated).

        :returns: Tuple with tag_updates and tag_conflicts.
            tag_updates is a dictionary with new tags, None is used for
            removed tags
            tag_conflicts is a set of tags that conflicted, each of which is
            (tagname, source_target, dest_target), or None if no copying was
            done.
        """
        intertags: InterTags = InterTags.get(self, to_tags)
        return intertags.merge(
            overwrite=overwrite, ignore_master=ignore_master,
            selector=selector)

    def set_tag(self, tag_name: str, revision: RevisionID) -> None:
        """Set a tag.

        :param tag_name: Tag name
        :param revision: Revision id
        :raise GhostTagsNotSupported: if revision is not present in
            the branch repository
        """
        raise NotImplementedError(self.set_tag)

    def lookup_tag(self, tag_name: str) -> RevisionID:
        """Look up a tag.

        :param tag_name: Tag to look up
        :raise NoSuchTag: Raised when tag does not exist
        :return: Matching revision id
        """
        raise NotImplementedError(self.lookup_tag)

    def delete_tag(self, tag_name: str) -> None:
        """Delete a tag.

        :param tag_name: Tag to delete
        :raise NoSuchTag: Raised when tag does not exist
        """
        raise NotImplementedError(self.delete_tag)

    def rename_revisions(
            self, rename_map: Dict[RevisionID, RevisionID]) -> None:
        """Rename revisions in this tags dictionary.

        :param rename_map: Dictionary mapping old revids to new revids
        """
        reverse_tags = self.get_reverse_tag_dict()
        for revid, names in reverse_tags.items():
            if revid in rename_map:
                for name in names:
                    self.set_tag(name, rename_map[revid])

    def has_tag(self, tag_name: str) -> bool:
        return tag_name in self.get_tag_dict()


class DisabledTags(Tags):
    """Tag storage that refuses to store anything.

    This is used by older formats that can't store tags.
    """

    def _not_supported(self, *a, **k):
        raise errors.TagsNotSupported(self.branch)

    set_tag = _not_supported
    get_tag_dict = _not_supported
    _set_tag_dict = _not_supported
    lookup_tag = _not_supported
    delete_tag = _not_supported

    def merge_to(self, to_tags, overwrite=False, ignore_master=False, selector=None):
        # we never have anything to copy
        return {}, []

    def rename_revisions(self, rename_map):
        # No tags, so nothing to rename
        pass

    def get_reverse_tag_dict(self):
        # There aren't any tags, so the reverse mapping is empty.
        return {}


class InterTags(InterObject[Tags]):
    """Operations between sets of tags.
    """

    _optimisers = []
    """The available optimised InterTags types."""

    @classmethod
    def is_compatible(klass, source: Tags, target: Tags) -> bool:
        # This is the default implementation
        return True

    def merge(self, overwrite: bool = False, ignore_master: bool = False,
              selector: Optional[TagSelector] = None
              ) -> Tuple[TagUpdates, Set[TagConflict]]:
        """Copy tags between repositories if necessary and possible.

        This method has common command-line behaviour about handling
        error cases.

        All new definitions are copied across, except that tags that already
        exist keep their existing definitions.

        :param to_tags: Branch to receive these tags
        :param overwrite: Overwrite conflicting tags in the target branch
        :param ignore_master: Do not modify the tags in the target's master
            branch (if any).  Default is false (so the master will be updated).
        :param selector: Callback that determines whether a tag should be
            copied. It should take a tag name and as argument and return a
            boolean.

        :returns: Tuple with tag_updates and tag_conflicts.
            tag_updates is a dictionary with new tags, None is used for
            removed tags
            tag_conflicts is a set of tags that conflicted, each of which is
            (tagname, source_target, dest_target), or None if no copying was
            done.
        """
        with contextlib.ExitStack() as stack:
            if self.source.branch == self.target.branch:
                return {}, set()
            if not self.source.branch.supports_tags():
                # obviously nothing to copy
                return {}, set()
            source_dict = self.source.get_tag_dict()
            if not source_dict:
                # no tags in the source, and we don't want to clobber anything
                # that's in the destination
                return {}, set()
            # We merge_to both master and child individually.
            #
            # It's possible for master and child to have differing sets of
            # tags, in which case it's possible to have different sets of
            # conflicts.  We report the union of both conflict sets.  In
            # that case it's likely the child and master have accepted
            # different tags from the source, which may be a surprising result, but
            # the best we can do in the circumstances.
            #
            # Ideally we'd improve this API to report the different conflicts
            # more clearly to the caller, but we don't want to break plugins
            # such as bzr-builddeb that use this API.
            stack.enter_context(self.target.branch.lock_write())
            if ignore_master:
                master = None
            else:
                master = self.target.branch.get_master_branch()
            if master is not None:
                stack.enter_context(master.lock_write())
            updates, conflicts = self._merge_to(
                self.target, source_dict, overwrite, selector=selector)
            if master is not None:
                extra_updates, extra_conflicts = self._merge_to(
                    master.tags, source_dict, overwrite, selector=selector)
                updates.update(extra_updates)
                conflicts += extra_conflicts
        # We use set() to remove any duplicate conflicts from the master
        # branch.
        return updates, set(conflicts)

    @classmethod
    def _merge_to(cls, to_tags, source_dict, overwrite, selector):
        dest_dict = to_tags.get_tag_dict()
        result, updates, conflicts = _reconcile_tags(
            source_dict, dest_dict, overwrite, selector)
        if result != dest_dict:
            to_tags._set_tag_dict(result)
        return updates, conflicts


class MemoryTags(Tags):

    def __init__(self, tag_dict):
        self._tag_dict = tag_dict

    def get_tag_dict(self):
        return self._tag_dict

    def lookup_tag(self, tag_name):
        """Return the referent string of a tag"""
        td = self.get_tag_dict()
        try:
            return td[tag_name]
        except KeyError:
            raise errors.NoSuchTag(tag_name)

    def set_tag(self, name, revid):
        self._tag_dict[name] = revid

    def delete_tag(self, name):
        try:
            del self._tag_dict[name]
        except KeyError:
            raise errors.NoSuchTag(name)

    def rename_revisions(self, revid_map):
        self._tag_dict = {
            name: revid_map.get(revid, revid)
            for name, revid in self._tag_dict.items()}

    def _set_tag_dict(self, result):
        self._tag_dict = dict(result.items())

    def merge_to(self, to_tags, overwrite=False, ignore_master=False, selector=None):
        source_dict = self.get_tag_dict()
        dest_dict = to_tags.get_tag_dict()
        result, updates, conflicts = _reconcile_tags(
            source_dict, dest_dict, overwrite, selector)
        if result != dest_dict:
            to_tags._set_tag_dict(result)
        return updates, conflicts


def sort_natural(branch, tags):
    """Sort tags, with numeric substrings as numbers.

    :param branch: Branch
    :param tags: List of tuples with tag name and revision id.
    """
    def natural_sort_key(tag):
        return [f(s) for f, s in
                zip(itertools.cycle((str.lower, int)),
                    re.split('([0-9]+)', tag[0]))]
    tags.sort(key=natural_sort_key)


def sort_alpha(branch, tags):
    """Sort tags lexicographically, in place.

    :param branch: Branch
    :param tags: List of tuples with tag name and revision id.
    """
    tags.sort()


def sort_time(branch, tags):
    """Sort tags by time inline.

    :param branch: Branch
    :param tags: List of tuples with tag name and revision id.
    """
    timestamps = {}
    for tag, revid in tags:
        try:
            revobj = branch.repository.get_revision(revid)
        except errors.NoSuchRevision:
            timestamp = sys.maxsize  # place them at the end
        else:
            timestamp = revobj.timestamp
        timestamps[revid] = timestamp
    tags.sort(key=lambda x: timestamps[x[1]])


tag_sort_methods = Registry[str, Callable[[_mod_branch.Branch, List[str]], List[str]]]()
tag_sort_methods.register("natural", sort_natural,
                          'Sort numeric substrings as numbers. (default)')
tag_sort_methods.register("alpha", sort_alpha, 'Sort tags lexicographically.')
tag_sort_methods.register("time", sort_time, 'Sort tags chronologically.')
tag_sort_methods.default_key = "natural"


InterTags.register_optimiser(InterTags)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""CVS working tree support.

Currently limited to referencing tools for migration.
"""

from ... import version_info  # noqa: F401
from ... import controldir, errors
from ...transport import register_transport_proto


class CVSUnsupportedError(errors.UnsupportedVcs):

    vcs = "cvs"

    _fmt = ("CVS working trees are not supported. To convert CVS projects to "
            "bzr, please see http://bazaar-vcs.org/BzrMigration and/or "
            "https://launchpad.net/launchpad-bazaar/+faq/26.")


class CVSDirFormat(controldir.ControlDirFormat):
    """The CVS directory control format."""

    def get_converter(self):
        raise NotImplementedError(self.get_converter)

    def get_format_description(self):
        return "CVS control directory."

    def initialize_on_transport(self, transport):
        raise errors.UninitializableFormat(self)

    def is_supported(self):
        return False

    def supports_transport(self, transport):
        return False

    def check_support_status(self, allow_unsupported, recommend_upgrade=True,
                             basedir=None):
        raise CVSUnsupportedError(format=self)

    def open(self, transport):
        # Raise NotBranchError if there is nothing there
        CVSProber().probe_transport(transport)
        raise NotImplementedError(self.open)


class CVSProber(controldir.Prober):

    @classmethod
    def priority(klass, transport):
        return 100

    @classmethod
    def probe_transport(klass, transport):
        # little ugly, but works
        # try a manual probe first, its a little faster perhaps ?
        if transport.has('CVS') and transport.has('CVS/Repository'):
            return CVSDirFormat()
        raise errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        return [CVSDirFormat()]


controldir.ControlDirFormat.register_prober(CVSProber)

register_transport_proto(
    'cvs+pserver://', help="The pserver access protocol for CVS.")



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from ... import errors, osutils, repository, revision, tests, workingtree
from ...tests.scenarios import load_tests_apply_scenarios
from .. import chk_map, groupcompress, inventory
from ..inventory import (ROOT_ID, CHKInventory, DuplicateFileId,
                         InvalidEntryName, Inventory, InventoryDirectory,
                         InventoryEntry, InventoryFile, TreeReference,
                         mutable_inventory_from_tree)
from . import TestCase, TestCaseWithTransport

load_tests = load_tests_apply_scenarios


def delta_application_scenarios():
    scenarios = [
        ('Inventory', {'apply_delta': apply_inventory_Inventory}),
        ]
    # Working tree basis delta application
    # Repository add_inv_by_delta.
    # Reduce form of the per_repository test logic - that logic needs to be
    # be able to get /just/ repositories whereas these tests are fine with
    # just creating trees.
    formats = set()
    for _, format in repository.format_registry.iteritems():
        if format.supports_full_versioned_files:
            scenarios.append((str(format.__name__), {
                'apply_delta': apply_inventory_Repository_add_inventory_by_delta,
                'format': format}))
    for getter in workingtree.format_registry._get_all_lazy():
        try:
            format = getter()
            if callable(format):
                format = format()
        except ImportError:
            pass  # Format with unmet dependency
        repo_fmt = format._matchingcontroldir.repository_format
        if not repo_fmt.supports_full_versioned_files:
            continue
        scenarios.append(
            (str(format.__class__.__name__) + ".update_basis_by_delta", {
                'apply_delta': apply_inventory_WT_basis,
                'format': format}))
        scenarios.append(
            (str(format.__class__.__name__) + ".apply_inventory_delta", {
                'apply_delta': apply_inventory_WT,
                'format': format}))
    return scenarios


def create_texts_for_inv(repo, inv):
    for path, ie in inv.iter_entries():
        if ie.text_size:
            lines = [b'a' * ie.text_size]
        else:
            lines = []
        repo.texts.add_lines((ie.file_id, ie.revision), [], lines)


def apply_inventory_Inventory(self, basis, delta, invalid_delta=True):
    """Apply delta to basis and return the result.

    :param basis: An inventory to be used as the basis.
    :param delta: The inventory delta to apply:
    :return: An inventory resulting from the application.
    """
    basis.apply_delta(delta)
    return basis


def apply_inventory_WT(self, basis, delta, invalid_delta=True):
    """Apply delta to basis and return the result.

    This sets the tree state to be basis, and then calls apply_inventory_delta.

    :param basis: An inventory to be used as the basis.
    :param delta: The inventory delta to apply:
    :return: An inventory resulting from the application.
    """
    control = self.make_controldir(
        'tree', format=self.format._matchingcontroldir)
    control.create_repository()
    control.create_branch()
    tree = self.format.initialize(control)
    tree.lock_write()
    try:
        tree._write_inventory(basis)
    finally:
        tree.unlock()
    # Fresh object, reads disk again.
    tree = tree.controldir.open_workingtree()
    tree.lock_write()
    try:
        tree.apply_inventory_delta(delta)
    finally:
        tree.unlock()
    # reload tree - ensure we get what was written.
    tree = tree.controldir.open_workingtree()
    tree.lock_read()
    self.addCleanup(tree.unlock)
    if not invalid_delta:
        tree._validate()
    return tree.root_inventory


def _create_repo_revisions(repo, basis, delta, invalid_delta):
    with repository.WriteGroup(repo):
        rev = revision.Revision(b'basis', timestamp=0, timezone=None,
                                message="", committer="foo@example.com")
        basis.revision_id = b'basis'
        create_texts_for_inv(repo, basis)
        repo.add_revision(b'basis', rev, basis)
        if invalid_delta:
            # We don't want to apply the delta to the basis, because we expect
            # the delta is invalid.
            result_inv = basis
            result_inv.revision_id = b'result'
            target_entries = None
        else:
            result_inv = basis.create_by_apply_delta(delta, b'result')
            create_texts_for_inv(repo, result_inv)
            target_entries = list(result_inv.iter_entries_by_dir())
        rev = revision.Revision(b'result', timestamp=0, timezone=None,
                                message="", committer="foo@example.com")
        repo.add_revision(b'result', rev, result_inv)
    return target_entries


def _get_basis_entries(tree):
    basis_tree = tree.basis_tree()
    with basis_tree.lock_read():
        return list(basis_tree.inventory.iter_entries_by_dir())


def _populate_different_tree(tree, basis, delta):
    """Put all entries into tree, but at a unique location."""
    added_ids = set()
    added_paths = set()
    tree.add(['unique-dir'], ['directory'], [b'unique-dir-id'])
    for path, ie in basis.iter_entries_by_dir():
        if ie.file_id in added_ids:
            continue
        # We want a unique path for each of these, we use the file-id
        tree.add(['unique-dir/' + ie.file_id], [ie.kind], [ie.file_id])
        added_ids.add(ie.file_id)
    for old_path, new_path, file_id, ie in delta:
        if file_id in added_ids:
            continue
        tree.add(['unique-dir/' + file_id], [ie.kind], [file_id])


def apply_inventory_WT_basis(test, basis, delta, invalid_delta=True):
    """Apply delta to basis and return the result.

    This sets the parent and then calls update_basis_by_delta.
    It also puts the basis in the repository under both 'basis' and 'result' to
    allow safety checks made by the WT to succeed, and finally ensures that all
    items in the delta with a new path are present in the WT before calling
    update_basis_by_delta.

    :param basis: An inventory to be used as the basis.
    :param delta: The inventory delta to apply:
    :return: An inventory resulting from the application.
    """
    control = test.make_controldir(
        'tree', format=test.format._matchingcontroldir)
    control.create_repository()
    control.create_branch()
    tree = test.format.initialize(control)
    tree.lock_write()
    try:
        target_entries = _create_repo_revisions(tree.branch.repository, basis,
                                                delta, invalid_delta)
        # Set the basis state as the trees current state
        tree._write_inventory(basis)
        # This reads basis from the repo and puts it into the tree's local
        # cache, if it has one.
        tree.set_parent_ids([b'basis'])
    finally:
        tree.unlock()
    # Fresh lock, reads disk again.
    with tree.lock_write():
        tree.update_basis_by_delta(b'result', delta)
        if not invalid_delta:
            tree._validate()
    # reload tree - ensure we get what was written.
    tree = tree.controldir.open_workingtree()
    basis_tree = tree.basis_tree()
    basis_tree.lock_read()
    test.addCleanup(basis_tree.unlock)
    basis_inv = basis_tree.root_inventory
    if target_entries:
        basis_entries = list(basis_inv.iter_entries_by_dir())
        test.assertEqual(target_entries, basis_entries)
    return basis_inv


def apply_inventory_Repository_add_inventory_by_delta(self, basis, delta,
                                                      invalid_delta=True):
    """Apply delta to basis and return the result.

    This inserts basis as a whole inventory and then uses
    add_inventory_by_delta to add delta.

    :param basis: An inventory to be used as the basis.
    :param delta: The inventory delta to apply:
    :return: An inventory resulting from the application.
    """
    format = self.format()
    control = self.make_controldir('tree', format=format._matchingcontroldir)
    repo = format.initialize(control)
    with repo.lock_write(), repository.WriteGroup(repo):
        rev = revision.Revision(
            b'basis', timestamp=0, timezone=None, message="",
            committer="foo@example.com")
        basis.revision_id = b'basis'
        create_texts_for_inv(repo, basis)
        repo.add_revision(b'basis', rev, basis)
    with repo.lock_write(), repository.WriteGroup(repo):
        inv_sha1 = repo.add_inventory_by_delta(
            b'basis', delta, b'result', [b'basis'])
    # Fresh lock, reads disk again.
    repo = repo.controldir.open_repository()
    repo.lock_read()
    self.addCleanup(repo.unlock)
    return repo.get_inventory(b'result')


class TestInventoryUpdates(TestCase):

    def test_creation_from_root_id(self):
        # iff a root id is passed to the constructor, a root directory is made
        inv = inventory.Inventory(root_id=b'tree-root')
        self.assertNotEqual(None, inv.root)
        self.assertEqual(b'tree-root', inv.root.file_id)

    def test_add_path_of_root(self):
        # if no root id is given at creation time, there is no root directory
        inv = inventory.Inventory(root_id=None)
        self.assertIs(None, inv.root)
        # add a root entry by adding its path
        ie = inv.add_path("", "directory", b"my-root")
        ie.revision = b'test-rev'
        self.assertEqual(b"my-root", ie.file_id)
        self.assertIs(ie, inv.root)

    def test_add_path(self):
        inv = inventory.Inventory(root_id=b'tree_root')
        ie = inv.add_path('hello', 'file', b'hello-id')
        self.assertEqual(b'hello-id', ie.file_id)
        self.assertEqual('file', ie.kind)

    def test_copy(self):
        """Make sure copy() works and creates a deep copy."""
        inv = inventory.Inventory(root_id=b'some-tree-root')
        ie = inv.add_path('hello', 'file', b'hello-id')
        inv2 = inv.copy()
        inv.root.file_id = b'some-new-root'
        ie.name = 'file2'
        self.assertEqual(b'some-tree-root', inv2.root.file_id)
        self.assertEqual('hello', inv2.get_entry(b'hello-id').name)

    def test_copy_empty(self):
        """Make sure an empty inventory can be copied."""
        inv = inventory.Inventory(root_id=None)
        inv2 = inv.copy()
        self.assertIs(None, inv2.root)

    def test_copy_copies_root_revision(self):
        """Make sure the revision of the root gets copied."""
        inv = inventory.Inventory(root_id=b'someroot')
        inv.root.revision = b'therev'
        inv2 = inv.copy()
        self.assertEqual(b'someroot', inv2.root.file_id)
        self.assertEqual(b'therev', inv2.root.revision)

    def test_create_tree_reference(self):
        inv = inventory.Inventory(b'tree-root-123')
        inv.add(TreeReference(
            b'nested-id', 'nested', parent_id=b'tree-root-123',
            revision=b'rev', reference_revision=b'rev2'))

    def test_error_encoding(self):
        inv = inventory.Inventory(b'tree-root')
        inv.add(InventoryFile(b'a-id', '\u1234', b'tree-root'))
        e = self.assertRaises(errors.InconsistentDelta, inv.add,
                              InventoryFile(b'b-id', '\u1234', b'tree-root'))
        self.assertContainsRe(str(e), '\\u1234')

    def test_add_recursive(self):
        parent = InventoryDirectory(b'src-id', 'src', b'tree-root')
        child = InventoryFile(b'hello-id', 'hello.c', b'src-id')
        parent.children[child.file_id] = child
        inv = inventory.Inventory(b'tree-root')
        inv.add(parent)
        self.assertEqual('src/hello.c', inv.id2path(b'hello-id'))


class TestDeltaApplication(TestCaseWithTransport):

    scenarios = delta_application_scenarios()

    def get_empty_inventory(self, reference_inv=None):
        """Get an empty inventory.

        Note that tests should not depend on the revision of the root for
        setting up test conditions, as it has to be flexible to accomodate non
        rich root repositories.

        :param reference_inv: If not None, get the revision for the root from
            this inventory. This is useful for dealing with older repositories
            that routinely discarded the root entry data. If None, the root's
            revision is set to 'basis'.
        """
        inv = inventory.Inventory()
        if reference_inv is not None:
            inv.root.revision = reference_inv.root.revision
        else:
            inv.root.revision = b'basis'
        return inv

    def make_file_ie(self, file_id=b'file-id', name='name', parent_id=None):
        ie_file = inventory.InventoryFile(file_id, name, parent_id)
        ie_file.revision = b'result'
        ie_file.text_size = 0
        ie_file.text_sha1 = b''
        return ie_file

    def test_empty_delta(self):
        inv = self.get_empty_inventory()
        delta = []
        inv = self.apply_delta(self, inv, delta)
        inv2 = self.get_empty_inventory(inv)
        self.assertEqual([], inv2._make_delta(inv))

    def test_None_file_id(self):
        inv = self.get_empty_inventory()
        dir1 = inventory.InventoryDirectory(b'dirid', 'dir1', inv.root.file_id)
        dir1.file_id = None
        dir1.revision = b'result'
        delta = [(None, 'dir1', None, dir1)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_unicode_file_id(self):
        inv = self.get_empty_inventory()
        dir1 = inventory.InventoryDirectory(b'dirid', 'dir1', inv.root.file_id)
        dir1.file_id = 'dirid'
        dir1.revision = b'result'
        delta = [(None, 'dir1', dir1.file_id, dir1)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_repeated_file_id(self):
        inv = self.get_empty_inventory()
        file1 = inventory.InventoryFile(b'id', 'path1', inv.root.file_id)
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        file2 = file1.copy()
        file2.name = 'path2'
        delta = [(None, 'path1', b'id', file1),
                 (None, 'path2', b'id', file2)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_repeated_new_path(self):
        inv = self.get_empty_inventory()
        file1 = inventory.InventoryFile(b'id1', 'path', inv.root.file_id)
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        file2 = file1.copy()
        file2.file_id = b'id2'
        delta = [(None, 'path', b'id1', file1),
                 (None, 'path', b'id2', file2)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_repeated_old_path(self):
        inv = self.get_empty_inventory()
        file1 = inventory.InventoryFile(b'id1', 'path', inv.root.file_id)
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        # We can't *create* a source inventory with the same path, but
        # a badly generated partial delta might claim the same source twice.
        # This would be buggy in two ways: the path is repeated in the delta,
        # And the path for one of the file ids doesn't match the source
        # location. Alternatively, we could have a repeated fileid, but that
        # is separately checked for.
        file2 = inventory.InventoryFile(b'id2', 'path2', inv.root.file_id)
        file2.revision = b'result'
        file2.text_size = 0
        file2.text_sha1 = b""
        inv.add(file1)
        inv.add(file2)
        delta = [('path', None, b'id1', None), ('path', None, b'id2', None)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_mismatched_id_entry_id(self):
        inv = self.get_empty_inventory()
        file1 = inventory.InventoryFile(b'id1', 'path', inv.root.file_id)
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        delta = [(None, 'path', b'id', file1)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_mismatched_new_path_entry_None(self):
        inv = self.get_empty_inventory()
        delta = [(None, 'path', b'id', None)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_mismatched_new_path_None_entry(self):
        inv = self.get_empty_inventory()
        file1 = inventory.InventoryFile(b'id1', 'path', inv.root.file_id)
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        delta = [("path", None, b'id1', file1)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_parent_is_not_directory(self):
        inv = self.get_empty_inventory()
        file1 = inventory.InventoryFile(b'id1', 'path', inv.root.file_id)
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        file2 = inventory.InventoryFile(b'id2', 'path2', b'id1')
        file2.revision = b'result'
        file2.text_size = 0
        file2.text_sha1 = b""
        inv.add(file1)
        delta = [(None, 'path/path2', b'id2', file2)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_parent_is_missing(self):
        inv = self.get_empty_inventory()
        file2 = inventory.InventoryFile(b'id2', 'path2', b'missingparent')
        file2.revision = b'result'
        file2.text_size = 0
        file2.text_sha1 = b""
        delta = [(None, 'path/path2', b'id2', file2)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_new_parent_path_has_wrong_id(self):
        inv = self.get_empty_inventory()
        parent1 = inventory.InventoryDirectory(b'p-1', 'dir', inv.root.file_id)
        parent1.revision = b'result'
        parent2 = inventory.InventoryDirectory(
            b'p-2', 'dir2', inv.root.file_id)
        parent2.revision = b'result'
        file1 = inventory.InventoryFile(b'id', 'path', b'p-2')
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        inv.add(parent1)
        inv.add(parent2)
        # This delta claims that file1 is at dir/path, but actually its at
        # dir2/path if you follow the inventory parent structure.
        delta = [(None, 'dir/path', b'id', file1)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_old_parent_path_is_wrong(self):
        inv = self.get_empty_inventory()
        parent1 = inventory.InventoryDirectory(b'p-1', 'dir', inv.root.file_id)
        parent1.revision = b'result'
        parent2 = inventory.InventoryDirectory(
            b'p-2', 'dir2', inv.root.file_id)
        parent2.revision = b'result'
        file1 = inventory.InventoryFile(b'id', 'path', b'p-2')
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        inv.add(parent1)
        inv.add(parent2)
        inv.add(file1)
        # This delta claims that file1 was at dir/path, but actually it was at
        # dir2/path if you follow the inventory parent structure.
        delta = [('dir/path', None, b'id', None)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_old_parent_path_is_for_other_id(self):
        inv = self.get_empty_inventory()
        parent1 = inventory.InventoryDirectory(b'p-1', 'dir', inv.root.file_id)
        parent1.revision = b'result'
        parent2 = inventory.InventoryDirectory(
            b'p-2', 'dir2', inv.root.file_id)
        parent2.revision = b'result'
        file1 = inventory.InventoryFile(b'id', 'path', b'p-2')
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b""
        file2 = inventory.InventoryFile(b'id2', 'path', b'p-1')
        file2.revision = b'result'
        file2.text_size = 0
        file2.text_sha1 = b""
        inv.add(parent1)
        inv.add(parent2)
        inv.add(file1)
        inv.add(file2)
        # This delta claims that file1 was at dir/path, but actually it was at
        # dir2/path if you follow the inventory parent structure. At dir/path
        # is another entry we should not delete.
        delta = [('dir/path', None, b'id', None)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_add_existing_id_new_path(self):
        inv = self.get_empty_inventory()
        parent1 = inventory.InventoryDirectory(
            b'p-1', 'dir1', inv.root.file_id)
        parent1.revision = b'result'
        parent2 = inventory.InventoryDirectory(
            b'p-1', 'dir2', inv.root.file_id)
        parent2.revision = b'result'
        inv.add(parent1)
        delta = [(None, 'dir2', b'p-1', parent2)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_add_new_id_existing_path(self):
        inv = self.get_empty_inventory()
        parent1 = inventory.InventoryDirectory(
            b'p-1', 'dir1', inv.root.file_id)
        parent1.revision = b'result'
        parent2 = inventory.InventoryDirectory(
            b'p-2', 'dir1', inv.root.file_id)
        parent2.revision = b'result'
        inv.add(parent1)
        delta = [(None, 'dir1', b'p-2', parent2)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_remove_dir_leaving_dangling_child(self):
        inv = self.get_empty_inventory()
        dir1 = inventory.InventoryDirectory(b'p-1', 'dir1', inv.root.file_id)
        dir1.revision = b'result'
        dir2 = inventory.InventoryDirectory(b'p-2', 'child1', b'p-1')
        dir2.revision = b'result'
        dir3 = inventory.InventoryDirectory(b'p-3', 'child2', b'p-1')
        dir3.revision = b'result'
        inv.add(dir1)
        inv.add(dir2)
        inv.add(dir3)
        delta = [('dir1', None, b'p-1', None),
                 ('dir1/child2', None, b'p-3', None)]
        self.assertRaises(errors.InconsistentDelta, self.apply_delta, self,
                          inv, delta)

    def test_add_file(self):
        inv = self.get_empty_inventory()
        file1 = inventory.InventoryFile(b'file-id', 'path', inv.root.file_id)
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b''
        delta = [(None, 'path', b'file-id', file1)]
        res_inv = self.apply_delta(self, inv, delta, invalid_delta=False)
        self.assertEqual(b'file-id', res_inv.get_entry(b'file-id').file_id)

    def test_remove_file(self):
        inv = self.get_empty_inventory()
        file1 = inventory.InventoryFile(b'file-id', 'path', inv.root.file_id)
        file1.revision = b'result'
        file1.text_size = 0
        file1.text_sha1 = b''
        inv.add(file1)
        delta = [('path', None, b'file-id', None)]
        res_inv = self.apply_delta(self, inv, delta, invalid_delta=False)
        self.assertEqual(None, res_inv.path2id('path'))
        self.assertRaises(errors.NoSuchId, res_inv.id2path, b'file-id')

    def test_rename_file(self):
        inv = self.get_empty_inventory()
        file1 = self.make_file_ie(name='path', parent_id=inv.root.file_id)
        inv.add(file1)
        file2 = self.make_file_ie(name='path2', parent_id=inv.root.file_id)
        delta = [('path', 'path2', b'file-id', file2)]
        res_inv = self.apply_delta(self, inv, delta, invalid_delta=False)
        self.assertEqual(None, res_inv.path2id('path'))
        self.assertEqual(b'file-id', res_inv.path2id('path2'))

    def test_replaced_at_new_path(self):
        inv = self.get_empty_inventory()
        file1 = self.make_file_ie(file_id=b'id1', parent_id=inv.root.file_id)
        inv.add(file1)
        file2 = self.make_file_ie(file_id=b'id2', parent_id=inv.root.file_id)
        delta = [('name', None, b'id1', None),
                 (None, 'name', b'id2', file2)]
        res_inv = self.apply_delta(self, inv, delta, invalid_delta=False)
        self.assertEqual(b'id2', res_inv.path2id('name'))

    def test_rename_dir(self):
        inv = self.get_empty_inventory()
        dir1 = inventory.InventoryDirectory(
            b'dir-id', 'dir1', inv.root.file_id)
        dir1.revision = b'basis'
        file1 = self.make_file_ie(parent_id=b'dir-id')
        inv.add(dir1)
        inv.add(file1)
        dir2 = inventory.InventoryDirectory(
            b'dir-id', 'dir2', inv.root.file_id)
        dir2.revision = b'result'
        delta = [('dir1', 'dir2', b'dir-id', dir2)]
        res_inv = self.apply_delta(self, inv, delta, invalid_delta=False)
        # The file should be accessible under the new path
        self.assertEqual(b'file-id', res_inv.path2id('dir2/name'))

    def test_renamed_dir_with_renamed_child(self):
        inv = self.get_empty_inventory()
        dir1 = inventory.InventoryDirectory(
            b'dir-id', 'dir1', inv.root.file_id)
        dir1.revision = b'basis'
        file1 = self.make_file_ie(b'file-id-1', 'name1', parent_id=b'dir-id')
        file2 = self.make_file_ie(b'file-id-2', 'name2', parent_id=b'dir-id')
        inv.add(dir1)
        inv.add(file1)
        inv.add(file2)
        dir2 = inventory.InventoryDirectory(
            b'dir-id', 'dir2', inv.root.file_id)
        dir2.revision = b'result'
        file2b = self.make_file_ie(b'file-id-2', 'name2', inv.root.file_id)
        delta = [('dir1', 'dir2', b'dir-id', dir2),
                 ('dir1/name2', 'name2', b'file-id-2', file2b)]
        res_inv = self.apply_delta(self, inv, delta, invalid_delta=False)
        # The file should be accessible under the new path
        self.assertEqual(b'file-id-1', res_inv.path2id('dir2/name1'))
        self.assertEqual(None, res_inv.path2id('dir2/name2'))
        self.assertEqual(b'file-id-2', res_inv.path2id('name2'))

    def test_is_root(self):
        """Ensure our root-checking code is accurate."""
        inv = inventory.Inventory(b'TREE_ROOT')
        self.assertTrue(inv.is_root(b'TREE_ROOT'))
        self.assertFalse(inv.is_root(b'booga'))
        inv.root.file_id = b'booga'
        self.assertFalse(inv.is_root(b'TREE_ROOT'))
        self.assertTrue(inv.is_root(b'booga'))
        # works properly even if no root is set
        inv.root = None
        self.assertFalse(inv.is_root(b'TREE_ROOT'))
        self.assertFalse(inv.is_root(b'booga'))

    def test_entries_for_empty_inventory(self):
        """Test that entries() will not fail for an empty inventory"""
        inv = Inventory(root_id=None)
        self.assertEqual([], inv.entries())


class TestInventoryEntry(TestCase):

    def test_file_invalid_entry_name(self):
        self.assertRaises(InvalidEntryName, inventory.InventoryFile,
                          b'123', 'a/hello.c', ROOT_ID)

    def test_file_backslash(self):
        file = inventory.InventoryFile(b'123', 'h\\ello.c', ROOT_ID)
        self.assertEqual(file.name, 'h\\ello.c')

    def test_file_kind_character(self):
        file = inventory.InventoryFile(b'123', 'hello.c', ROOT_ID)
        self.assertEqual(file.kind_character(), '')

    def test_dir_kind_character(self):
        dir = inventory.InventoryDirectory(b'123', 'hello.c', ROOT_ID)
        self.assertEqual(dir.kind_character(), '/')

    def test_link_kind_character(self):
        dir = inventory.InventoryLink(b'123', 'hello.c', ROOT_ID)
        self.assertEqual(dir.kind_character(), '')

    def test_tree_ref_kind_character(self):
        dir = TreeReference(b'123', 'hello.c', ROOT_ID)
        self.assertEqual(dir.kind_character(), '+')

    def test_dir_detect_changes(self):
        left = inventory.InventoryDirectory(b'123', 'hello.c', ROOT_ID)
        right = inventory.InventoryDirectory(b'123', 'hello.c', ROOT_ID)
        self.assertEqual((False, False), left.detect_changes(right))
        self.assertEqual((False, False), right.detect_changes(left))

    def test_file_detect_changes(self):
        left = inventory.InventoryFile(b'123', 'hello.c', ROOT_ID)
        left.text_sha1 = 123
        right = inventory.InventoryFile(b'123', 'hello.c', ROOT_ID)
        right.text_sha1 = 123
        self.assertEqual((False, False), left.detect_changes(right))
        self.assertEqual((False, False), right.detect_changes(left))
        left.executable = True
        self.assertEqual((False, True), left.detect_changes(right))
        self.assertEqual((False, True), right.detect_changes(left))
        right.text_sha1 = 321
        self.assertEqual((True, True), left.detect_changes(right))
        self.assertEqual((True, True), right.detect_changes(left))

    def test_symlink_detect_changes(self):
        left = inventory.InventoryLink(b'123', 'hello.c', ROOT_ID)
        left.symlink_target = 'foo'
        right = inventory.InventoryLink(b'123', 'hello.c', ROOT_ID)
        right.symlink_target = 'foo'
        self.assertEqual((False, False), left.detect_changes(right))
        self.assertEqual((False, False), right.detect_changes(left))
        left.symlink_target = 'different'
        self.assertEqual((True, False), left.detect_changes(right))
        self.assertEqual((True, False), right.detect_changes(left))

    def test_file_has_text(self):
        file = inventory.InventoryFile(b'123', 'hello.c', ROOT_ID)
        self.assertTrue(file.has_text())

    def test_directory_has_text(self):
        dir = inventory.InventoryDirectory(b'123', 'hello.c', ROOT_ID)
        self.assertFalse(dir.has_text())

    def test_link_has_text(self):
        link = inventory.InventoryLink(b'123', 'hello.c', ROOT_ID)
        self.assertFalse(link.has_text())

    def test_make_entry(self):
        self.assertIsInstance(inventory.make_entry("file", "name", ROOT_ID),
                              inventory.InventoryFile)
        self.assertIsInstance(inventory.make_entry("symlink", "name", ROOT_ID),
                              inventory.InventoryLink)
        self.assertIsInstance(inventory.make_entry("directory", "name", ROOT_ID),
                              inventory.InventoryDirectory)

    def test_make_entry_non_normalized(self):
        orig_normalized_filename = osutils.normalized_filename

        try:
            osutils.normalized_filename = osutils._accessible_normalized_filename
            entry = inventory.make_entry("file", 'a\u030a', ROOT_ID)
            self.assertEqual('\xe5', entry.name)
            self.assertIsInstance(entry, inventory.InventoryFile)

            osutils.normalized_filename = osutils._inaccessible_normalized_filename
            self.assertRaises(errors.InvalidNormalization,
                              inventory.make_entry, 'file', 'a\u030a', ROOT_ID)
        finally:
            osutils.normalized_filename = orig_normalized_filename


class TestDescribeChanges(TestCase):

    def test_describe_change(self):
        # we need to test the following change combinations:
        # rename
        # reparent
        # modify
        # gone
        # added
        # renamed/reparented and modified
        # change kind (perhaps can't be done yet?)
        # also, merged in combination with all of these?
        old_a = InventoryFile(b'a-id', 'a_file', ROOT_ID)
        old_a.text_sha1 = b'123132'
        old_a.text_size = 0
        new_a = InventoryFile(b'a-id', 'a_file', ROOT_ID)
        new_a.text_sha1 = b'123132'
        new_a.text_size = 0

        self.assertChangeDescription('unchanged', old_a, new_a)

        new_a.text_size = 10
        new_a.text_sha1 = b'abcabc'
        self.assertChangeDescription('modified', old_a, new_a)

        self.assertChangeDescription('added', None, new_a)
        self.assertChangeDescription('removed', old_a, None)
        # perhaps a bit questionable but seems like the most reasonable thing...
        self.assertChangeDescription('unchanged', None, None)

        # in this case it's both renamed and modified; show a rename and
        # modification:
        new_a.name = 'newfilename'
        self.assertChangeDescription('modified and renamed', old_a, new_a)

        # reparenting is 'renaming'
        new_a.name = old_a.name
        new_a.parent_id = b'somedir-id'
        self.assertChangeDescription('modified and renamed', old_a, new_a)

        # reset the content values so its not modified
        new_a.text_size = old_a.text_size
        new_a.text_sha1 = old_a.text_sha1
        new_a.name = old_a.name

        new_a.name = 'newfilename'
        self.assertChangeDescription('renamed', old_a, new_a)

        # reparenting is 'renaming'
        new_a.name = old_a.name
        new_a.parent_id = b'somedir-id'
        self.assertChangeDescription('renamed', old_a, new_a)

    def assertChangeDescription(self, expected_change, old_ie, new_ie):
        change = InventoryEntry.describe_change(old_ie, new_ie)
        self.assertEqual(expected_change, change)


class TestCHKInventory(tests.TestCaseWithMemoryTransport):

    def get_chk_bytes(self):
        factory = groupcompress.make_pack_factory(True, True, 1)
        trans = self.get_transport('')
        return factory(trans)

    def read_bytes(self, chk_bytes, key):
        stream = chk_bytes.get_record_stream([key], 'unordered', True)
        return next(stream).get_bytes_as("fulltext")

    def test_deserialise_gives_CHKInventory(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertEqual(b"revid", new_inv.revision_id)
        self.assertEqual("directory", new_inv.root.kind)
        self.assertEqual(inv.root.file_id, new_inv.root.file_id)
        self.assertEqual(inv.root.parent_id, new_inv.root.parent_id)
        self.assertEqual(inv.root.name, new_inv.root.name)
        self.assertEqual(b"rootrev", new_inv.root.revision)
        self.assertEqual(b'plain', new_inv._search_key_name)

    def test_deserialise_wrong_revid(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        self.assertRaises(ValueError, CHKInventory.deserialise, chk_bytes,
                          lines, (b"revid2",))

    def test_captures_rev_root_byid(self):
        inv = Inventory()
        inv.revision_id = b"foo"
        inv.root.revision = b"bar"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        self.assertEqual([
            b'chkinventory:\n',
            b'revision_id: foo\n',
            b'root_id: TREE_ROOT\n',
            b'parent_id_basename_to_file_id: sha1:eb23f0ad4b07f48e88c76d4c94292be57fb2785f\n',
            b'id_to_entry: sha1:debfe920f1f10e7929260f0534ac9a24d7aabbb4\n',
            ], lines)
        chk_inv = CHKInventory.deserialise(
            chk_bytes, lines, (b'foo',))
        self.assertEqual(b'plain', chk_inv._search_key_name)

    def test_captures_parent_id_basename_index(self):
        inv = Inventory()
        inv.revision_id = b"foo"
        inv.root.revision = b"bar"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        self.assertEqual([
            b'chkinventory:\n',
            b'revision_id: foo\n',
            b'root_id: TREE_ROOT\n',
            b'parent_id_basename_to_file_id: sha1:eb23f0ad4b07f48e88c76d4c94292be57fb2785f\n',
            b'id_to_entry: sha1:debfe920f1f10e7929260f0534ac9a24d7aabbb4\n',
            ], lines)
        chk_inv = CHKInventory.deserialise(
            chk_bytes, lines, (b'foo',))
        self.assertEqual(b'plain', chk_inv._search_key_name)

    def test_captures_search_key_name(self):
        inv = Inventory()
        inv.revision_id = b"foo"
        inv.root.revision = b"bar"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv,
                                              search_key_name=b'hash-16-way')
        lines = chk_inv.to_lines()
        self.assertEqual([
            b'chkinventory:\n',
            b'search_key_name: hash-16-way\n',
            b'root_id: TREE_ROOT\n',
            b'parent_id_basename_to_file_id: sha1:eb23f0ad4b07f48e88c76d4c94292be57fb2785f\n',
            b'revision_id: foo\n',
            b'id_to_entry: sha1:debfe920f1f10e7929260f0534ac9a24d7aabbb4\n',
            ], lines)
        chk_inv = CHKInventory.deserialise(
            chk_bytes, lines, (b'foo',))
        self.assertEqual(b'hash-16-way', chk_inv._search_key_name)

    def test_directory_children_on_demand(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        inv.add(InventoryFile(b"fileid", "file", inv.root.file_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        root_entry = new_inv.get_entry(inv.root.file_id)
        self.assertEqual(None, root_entry._children)
        self.assertEqual({'file'}, set(root_entry.children))
        file_direct = new_inv.get_entry(b"fileid")
        file_found = root_entry.children['file']
        self.assertEqual(file_direct.kind, file_found.kind)
        self.assertEqual(file_direct.file_id, file_found.file_id)
        self.assertEqual(file_direct.parent_id, file_found.parent_id)
        self.assertEqual(file_direct.name, file_found.name)
        self.assertEqual(file_direct.revision, file_found.revision)
        self.assertEqual(file_direct.text_sha1, file_found.text_sha1)
        self.assertEqual(file_direct.text_size, file_found.text_size)
        self.assertEqual(file_direct.executable, file_found.executable)

    def test_from_inventory_maximum_size(self):
        # from_inventory supports the maximum_size parameter.
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv, 120)
        chk_inv.id_to_entry._ensure_root()
        self.assertEqual(120, chk_inv.id_to_entry._root_node.maximum_size)
        self.assertEqual(1, chk_inv.id_to_entry._root_node._key_width)
        p_id_basename = chk_inv.parent_id_basename_to_file_id
        p_id_basename._ensure_root()
        self.assertEqual(120, p_id_basename._root_node.maximum_size)
        self.assertEqual(2, p_id_basename._root_node._key_width)

    def test_iter_all_ids(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        inv.add(InventoryFile(b"fileid", "file", inv.root.file_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        fileids = sorted(new_inv.iter_all_ids())
        self.assertEqual([inv.root.file_id, b"fileid"], fileids)

    def test__len__(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        inv.add(InventoryFile(b"fileid", "file", inv.root.file_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        self.assertEqual(2, len(chk_inv))

    def test_get_entry(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        inv.add(InventoryFile(b"fileid", "file", inv.root.file_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        root_entry = new_inv.get_entry(inv.root.file_id)
        file_entry = new_inv.get_entry(b"fileid")
        self.assertEqual("directory", root_entry.kind)
        self.assertEqual(inv.root.file_id, root_entry.file_id)
        self.assertEqual(inv.root.parent_id, root_entry.parent_id)
        self.assertEqual(inv.root.name, root_entry.name)
        self.assertEqual(b"rootrev", root_entry.revision)
        self.assertEqual("file", file_entry.kind)
        self.assertEqual(b"fileid", file_entry.file_id)
        self.assertEqual(inv.root.file_id, file_entry.parent_id)
        self.assertEqual("file", file_entry.name)
        self.assertEqual(b"filerev", file_entry.revision)
        self.assertEqual(b"ffff", file_entry.text_sha1)
        self.assertEqual(1, file_entry.text_size)
        self.assertEqual(True, file_entry.executable)
        self.assertRaises(errors.NoSuchId, new_inv.get_entry, 'missing')

    def test_has_id_true(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        inv.add(InventoryFile(b"fileid", "file", inv.root.file_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        self.assertTrue(chk_inv.has_id(b'fileid'))
        self.assertTrue(chk_inv.has_id(inv.root.file_id))

    def test_has_id_not(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        self.assertFalse(chk_inv.has_id(b'fileid'))

    def test_id2path(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        direntry = InventoryDirectory(b"dirid", "dir", inv.root.file_id)
        fileentry = InventoryFile(b"fileid", "file", b"dirid")
        inv.add(direntry)
        inv.add(fileentry)
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        inv.get_entry(b"dirid").revision = b"filerev"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertEqual('', new_inv.id2path(inv.root.file_id))
        self.assertEqual('dir', new_inv.id2path(b'dirid'))
        self.assertEqual('dir/file', new_inv.id2path(b'fileid'))

    def test_path2id(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        direntry = InventoryDirectory(b"dirid", "dir", inv.root.file_id)
        fileentry = InventoryFile(b"fileid", "file", b"dirid")
        inv.add(direntry)
        inv.add(fileentry)
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        inv.get_entry(b"dirid").revision = b"filerev"
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertEqual(inv.root.file_id, new_inv.path2id(''))
        self.assertEqual(b'dirid', new_inv.path2id('dir'))
        self.assertEqual(b'fileid', new_inv.path2id('dir/file'))

    def test_create_by_apply_delta_sets_root(self):
        inv = Inventory()
        inv.root.revision = b"myrootrev"
        inv.revision_id = b"revid"
        chk_bytes = self.get_chk_bytes()
        base_inv = CHKInventory.from_inventory(chk_bytes, inv)
        inv.add_path("", "directory", b"myrootid", None)
        inv.revision_id = b"expectedid"
        inv.root.revision = b"myrootrev"
        reference_inv = CHKInventory.from_inventory(chk_bytes, inv)
        delta = [("", None, base_inv.root.file_id, None),
                 (None, "", b"myrootid", inv.root)]
        new_inv = base_inv.create_by_apply_delta(delta, b"expectedid")
        self.assertEqual(reference_inv.root, new_inv.root)

    def test_create_by_apply_delta_empty_add_child(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        chk_bytes = self.get_chk_bytes()
        base_inv = CHKInventory.from_inventory(chk_bytes, inv)
        a_entry = InventoryFile(b"A-id", "A", inv.root.file_id)
        a_entry.revision = b"filerev"
        a_entry.executable = True
        a_entry.text_sha1 = b"ffff"
        a_entry.text_size = 1
        inv.add(a_entry)
        inv.revision_id = b"expectedid"
        reference_inv = CHKInventory.from_inventory(chk_bytes, inv)
        delta = [(None, "A", b"A-id", a_entry)]
        new_inv = base_inv.create_by_apply_delta(delta, b"expectedid")
        # new_inv should be the same as reference_inv.
        self.assertEqual(reference_inv.revision_id, new_inv.revision_id)
        self.assertEqual(reference_inv.root_id, new_inv.root_id)
        reference_inv.id_to_entry._ensure_root()
        new_inv.id_to_entry._ensure_root()
        self.assertEqual(reference_inv.id_to_entry._root_node._key,
                         new_inv.id_to_entry._root_node._key)

    def test_create_by_apply_delta_empty_add_child_updates_parent_id(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        chk_bytes = self.get_chk_bytes()
        base_inv = CHKInventory.from_inventory(chk_bytes, inv)
        a_entry = InventoryFile(b"A-id", "A", inv.root.file_id)
        a_entry.revision = b"filerev"
        a_entry.executable = True
        a_entry.text_sha1 = b"ffff"
        a_entry.text_size = 1
        inv.add(a_entry)
        inv.revision_id = b"expectedid"
        reference_inv = CHKInventory.from_inventory(chk_bytes, inv)
        delta = [(None, "A", b"A-id", a_entry)]
        new_inv = base_inv.create_by_apply_delta(delta, b"expectedid")
        reference_inv.id_to_entry._ensure_root()
        reference_inv.parent_id_basename_to_file_id._ensure_root()
        new_inv.id_to_entry._ensure_root()
        new_inv.parent_id_basename_to_file_id._ensure_root()
        # new_inv should be the same as reference_inv.
        self.assertEqual(reference_inv.revision_id, new_inv.revision_id)
        self.assertEqual(reference_inv.root_id, new_inv.root_id)
        self.assertEqual(reference_inv.id_to_entry._root_node._key,
                         new_inv.id_to_entry._root_node._key)
        self.assertEqual(reference_inv.parent_id_basename_to_file_id._root_node._key,
                         new_inv.parent_id_basename_to_file_id._root_node._key)

    def test_iter_changes(self):
        # Low level bootstrapping smoke test; comprehensive generic tests via
        # InterTree are coming.
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        inv.add(InventoryFile(b"fileid", "file", inv.root.file_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        inv2 = Inventory()
        inv2.revision_id = b"revid2"
        inv2.root.revision = b"rootrev"
        inv2.add(InventoryFile(b"fileid", "file", inv.root.file_id))
        inv2.get_entry(b"fileid").revision = b"filerev2"
        inv2.get_entry(b"fileid").executable = False
        inv2.get_entry(b"fileid").text_sha1 = b"bbbb"
        inv2.get_entry(b"fileid").text_size = 2
        # get fresh objects.
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        inv_1 = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        chk_inv2 = CHKInventory.from_inventory(chk_bytes, inv2)
        lines = chk_inv2.to_lines()
        inv_2 = CHKInventory.deserialise(chk_bytes, lines, (b"revid2",))
        self.assertEqual([(b'fileid', ('file', 'file'), True, (True, True),
                           (b'TREE_ROOT', b'TREE_ROOT'), ('file',
                                                          'file'), ('file', 'file'),
                           (False, True))],
                         list(inv_1.iter_changes(inv_2)))

    def test_parent_id_basename_to_file_id_index_enabled(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        inv.add(InventoryFile(b"fileid", "file", inv.root.file_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        # get fresh objects.
        chk_bytes = self.get_chk_bytes()
        tmp_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = tmp_inv.to_lines()
        chk_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertIsInstance(
            chk_inv.parent_id_basename_to_file_id, chk_map.CHKMap)
        self.assertEqual(
            {(b'', b''): b'TREE_ROOT', (b'TREE_ROOT', b'file'): b'fileid'},
            dict(chk_inv.parent_id_basename_to_file_id.iteritems()))

    def test_file_entry_to_bytes(self):
        inv = CHKInventory(None)
        ie = inventory.InventoryFile(b'file-id', 'filename', b'parent-id')
        ie.executable = True
        ie.revision = b'file-rev-id'
        ie.text_sha1 = b'abcdefgh'
        ie.text_size = 100
        bytes = inv._entry_to_bytes(ie)
        self.assertEqual(b'file: file-id\nparent-id\nfilename\n'
                         b'file-rev-id\nabcdefgh\n100\nY', bytes)
        ie2 = inv._bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual((b'filename', b'file-id', b'file-rev-id'),
                         inv._bytes_to_utf8name_key(bytes))

    def test_file2_entry_to_bytes(self):
        inv = CHKInventory(None)
        # \u30a9 == 'omega'
        ie = inventory.InventoryFile(b'file-id', '\u03a9name', b'parent-id')
        ie.executable = False
        ie.revision = b'file-rev-id'
        ie.text_sha1 = b'123456'
        ie.text_size = 25
        bytes = inv._entry_to_bytes(ie)
        self.assertEqual(b'file: file-id\nparent-id\n\xce\xa9name\n'
                         b'file-rev-id\n123456\n25\nN', bytes)
        ie2 = inv._bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual((b'\xce\xa9name', b'file-id', b'file-rev-id'),
                         inv._bytes_to_utf8name_key(bytes))

    def test_dir_entry_to_bytes(self):
        inv = CHKInventory(None)
        ie = inventory.InventoryDirectory(b'dir-id', 'dirname', b'parent-id')
        ie.revision = b'dir-rev-id'
        bytes = inv._entry_to_bytes(ie)
        self.assertEqual(b'dir: dir-id\nparent-id\ndirname\ndir-rev-id', bytes)
        ie2 = inv._bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual((b'dirname', b'dir-id', b'dir-rev-id'),
                         inv._bytes_to_utf8name_key(bytes))

    def test_dir2_entry_to_bytes(self):
        inv = CHKInventory(None)
        ie = inventory.InventoryDirectory(b'dir-id', 'dir\u03a9name',
                                          None)
        ie.revision = b'dir-rev-id'
        bytes = inv._entry_to_bytes(ie)
        self.assertEqual(b'dir: dir-id\n\ndir\xce\xa9name\n'
                         b'dir-rev-id', bytes)
        ie2 = inv._bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertIs(ie2.parent_id, None)
        self.assertEqual((b'dir\xce\xa9name', b'dir-id', b'dir-rev-id'),
                         inv._bytes_to_utf8name_key(bytes))

    def test_symlink_entry_to_bytes(self):
        inv = CHKInventory(None)
        ie = inventory.InventoryLink(b'link-id', 'linkname', b'parent-id')
        ie.revision = b'link-rev-id'
        ie.symlink_target = 'target/path'
        bytes = inv._entry_to_bytes(ie)
        self.assertEqual(b'symlink: link-id\nparent-id\nlinkname\n'
                         b'link-rev-id\ntarget/path', bytes)
        ie2 = inv._bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertIsInstance(ie2.symlink_target, str)
        self.assertEqual((b'linkname', b'link-id', b'link-rev-id'),
                         inv._bytes_to_utf8name_key(bytes))

    def test_symlink2_entry_to_bytes(self):
        inv = CHKInventory(None)
        ie = inventory.InventoryLink(
            b'link-id', 'link\u03a9name', b'parent-id')
        ie.revision = b'link-rev-id'
        ie.symlink_target = 'target/\u03a9path'
        bytes = inv._entry_to_bytes(ie)
        self.assertEqual(b'symlink: link-id\nparent-id\nlink\xce\xa9name\n'
                         b'link-rev-id\ntarget/\xce\xa9path', bytes)
        ie2 = inv._bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertIsInstance(ie2.symlink_target, str)
        self.assertEqual((b'link\xce\xa9name', b'link-id', b'link-rev-id'),
                         inv._bytes_to_utf8name_key(bytes))

    def test_tree_reference_entry_to_bytes(self):
        inv = CHKInventory(None)
        ie = inventory.TreeReference(b'tree-root-id', 'tree\u03a9name',
                                     b'parent-id')
        ie.revision = b'tree-rev-id'
        ie.reference_revision = b'ref-rev-id'
        bytes = inv._entry_to_bytes(ie)
        self.assertEqual(b'tree: tree-root-id\nparent-id\ntree\xce\xa9name\n'
                         b'tree-rev-id\nref-rev-id', bytes)
        ie2 = inv._bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual((b'tree\xce\xa9name', b'tree-root-id', b'tree-rev-id'),
                         inv._bytes_to_utf8name_key(bytes))

    def make_basic_utf8_inventory(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        root_id = inv.root.file_id
        inv.add(InventoryFile(b"fileid", 'f\xefle', root_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 0
        inv.add(InventoryDirectory(b"dirid", 'dir-\N{EURO SIGN}', root_id))
        inv.get_entry(b"dirid").revision = b"dirrev"
        inv.add(InventoryFile(b"childid", 'ch\xefld', b"dirid"))
        inv.get_entry(b"childid").revision = b"filerev"
        inv.get_entry(b"childid").text_sha1 = b"ffff"
        inv.get_entry(b"childid").text_size = 0
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        return CHKInventory.deserialise(chk_bytes, lines, (b"revid",))

    def test__preload_handles_utf8(self):
        new_inv = self.make_basic_utf8_inventory()
        self.assertEqual({}, new_inv._fileid_to_entry_cache)
        self.assertFalse(new_inv._fully_cached)
        new_inv._preload_cache()
        self.assertEqual(
            sorted([new_inv.root_id, b"fileid", b"dirid", b"childid"]),
            sorted(new_inv._fileid_to_entry_cache.keys()))
        ie_root = new_inv._fileid_to_entry_cache[new_inv.root_id]
        self.assertEqual(['dir-\N{EURO SIGN}', 'f\xefle'],
                         sorted(ie_root._children.keys()))
        ie_dir = new_inv._fileid_to_entry_cache[b'dirid']
        self.assertEqual(['ch\xefld'], sorted(ie_dir._children.keys()))

    def test__preload_populates_cache(self):
        inv = Inventory()
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        root_id = inv.root.file_id
        inv.add(InventoryFile(b"fileid", "file", root_id))
        inv.get_entry(b"fileid").revision = b"filerev"
        inv.get_entry(b"fileid").executable = True
        inv.get_entry(b"fileid").text_sha1 = b"ffff"
        inv.get_entry(b"fileid").text_size = 1
        inv.add(InventoryDirectory(b"dirid", "dir", root_id))
        inv.get_entry(b"dirid").revision = b"dirrev"
        inv.add(InventoryFile(b"childid", "child", b"dirid"))
        inv.get_entry(b"childid").revision = b"filerev"
        inv.get_entry(b"childid").executable = False
        inv.get_entry(b"childid").text_sha1 = b"dddd"
        inv.get_entry(b"childid").text_size = 1
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertEqual({}, new_inv._fileid_to_entry_cache)
        self.assertFalse(new_inv._fully_cached)
        new_inv._preload_cache()
        self.assertEqual(
            sorted([root_id, b"fileid", b"dirid", b"childid"]),
            sorted(new_inv._fileid_to_entry_cache.keys()))
        self.assertTrue(new_inv._fully_cached)
        ie_root = new_inv._fileid_to_entry_cache[root_id]
        self.assertEqual(['dir', 'file'], sorted(ie_root._children.keys()))
        ie_dir = new_inv._fileid_to_entry_cache[b'dirid']
        self.assertEqual(['child'], sorted(ie_dir._children.keys()))

    def test__preload_handles_partially_evaluated_inventory(self):
        new_inv = self.make_basic_utf8_inventory()
        ie = new_inv.get_entry(new_inv.root_id)
        self.assertIs(None, ie._children)
        self.assertEqual(['dir-\N{EURO SIGN}', 'f\xefle'],
                         sorted(ie.children.keys()))
        # Accessing .children loads _children
        self.assertEqual(['dir-\N{EURO SIGN}', 'f\xefle'],
                         sorted(ie._children.keys()))
        new_inv._preload_cache()
        # No change
        self.assertEqual(['dir-\N{EURO SIGN}', 'f\xefle'],
                         sorted(ie._children.keys()))
        ie_dir = new_inv.get_entry(b"dirid")
        self.assertEqual(['ch\xefld'],
                         sorted(ie_dir._children.keys()))

    def test_filter_change_in_renamed_subfolder(self):
        inv = Inventory(b'tree-root')
        inv.root.revision = b'rootrev'
        src_ie = inv.add_path('src', 'directory', b'src-id')
        src_ie.revision = b'srcrev'
        sub_ie = inv.add_path('src/sub/', 'directory', b'sub-id')
        sub_ie.revision = b'subrev'
        a_ie = inv.add_path('src/sub/a', 'file', b'a-id')
        a_ie.revision = b'filerev'
        a_ie.text_sha1 = osutils.sha_string(b'content\n')
        a_ie.text_size = len(b'content\n')
        chk_bytes = self.get_chk_bytes()
        inv = CHKInventory.from_inventory(chk_bytes, inv)
        inv = inv.create_by_apply_delta([
            ("src/sub/a", "src/sub/a", b"a-id", a_ie),
            ("src", "src2", b"src-id", src_ie),
            ], b'new-rev-2')
        new_inv = inv.filter([b'a-id', b'src-id'])
        self.assertEqual([
            ('', b'tree-root'),
            ('src', b'src-id'),
            ('src/sub', b'sub-id'),
            ('src/sub/a', b'a-id'),
            ], [(path, ie.file_id) for path, ie in new_inv.iter_entries()])


class TestCHKInventoryExpand(tests.TestCaseWithMemoryTransport):

    def get_chk_bytes(self):
        factory = groupcompress.make_pack_factory(True, True, 1)
        trans = self.get_transport('')
        return factory(trans)

    def make_dir(self, inv, name, parent_id, revision):
        ie = inv.make_entry('directory', name, parent_id,
                            name.encode('utf-8') + b'-id')
        ie.revision = revision
        inv.add(ie)

    def make_file(self, inv, name, parent_id, revision, content=b'content\n'):
        ie = inv.make_entry('file', name, parent_id,
                            name.encode('utf-8') + b'-id')
        ie.text_sha1 = osutils.sha_string(content)
        ie.text_size = len(content)
        ie.revision = revision
        inv.add(ie)

    def make_simple_inventory(self):
        inv = Inventory(b'TREE_ROOT')
        inv.revision_id = b"revid"
        inv.root.revision = b"rootrev"
        # /                 TREE_ROOT
        # dir1/             dir1-id
        #   sub-file1       sub-file1-id
        #   sub-file2       sub-file2-id
        #   sub-dir1/       sub-dir1-id
        #     subsub-file1  subsub-file1-id
        # dir2/             dir2-id
        #   sub2-file1      sub2-file1-id
        # top               top-id
        self.make_dir(inv, 'dir1', b'TREE_ROOT', b'dirrev')
        self.make_dir(inv, 'dir2', b'TREE_ROOT', b'dirrev')
        self.make_dir(inv, 'sub-dir1', b'dir1-id', b'dirrev')
        self.make_file(inv, 'top', b'TREE_ROOT', b'filerev')
        self.make_file(inv, 'sub-file1', b'dir1-id', b'filerev')
        self.make_file(inv, 'sub-file2', b'dir1-id', b'filerev')
        self.make_file(inv, 'subsub-file1', b'sub-dir1-id', b'filerev')
        self.make_file(inv, 'sub2-file1', b'dir2-id', b'filerev')
        chk_bytes = self.get_chk_bytes()
        #  use a small maximum_size to force internal paging structures
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv,
                                              maximum_size=100,
                                              search_key_name=b'hash-255-way')
        lines = chk_inv.to_lines()
        return CHKInventory.deserialise(chk_bytes, lines, (b"revid",))

    def assert_Getitems(self, expected_fileids, inv, file_ids):
        self.assertEqual(sorted(expected_fileids),
                         sorted([ie.file_id for ie in inv._getitems(file_ids)]))

    def assertExpand(self, all_ids, inv, file_ids):
        (val_all_ids,
         val_children) = inv._expand_fileids_to_parents_and_children(file_ids)
        self.assertEqual(set(all_ids), val_all_ids)
        entries = inv._getitems(val_all_ids)
        expected_children = {}
        for entry in entries:
            s = expected_children.setdefault(entry.parent_id, [])
            s.append(entry.file_id)
        val_children = {
            k: sorted(v) for k, v in val_children.items()}
        expected_children = {
            k: sorted(v) for k, v in expected_children.items()}
        self.assertEqual(expected_children, val_children)

    def test_make_simple_inventory(self):
        inv = self.make_simple_inventory()
        layout = []
        for path, entry in inv.iter_entries_by_dir():
            layout.append((path, entry.file_id))
        self.assertEqual([
            ('', b'TREE_ROOT'),
            ('dir1', b'dir1-id'),
            ('dir2', b'dir2-id'),
            ('top', b'top-id'),
            ('dir1/sub-dir1', b'sub-dir1-id'),
            ('dir1/sub-file1', b'sub-file1-id'),
            ('dir1/sub-file2', b'sub-file2-id'),
            ('dir1/sub-dir1/subsub-file1', b'subsub-file1-id'),
            ('dir2/sub2-file1', b'sub2-file1-id'),
            ], layout)

    def test__getitems(self):
        inv = self.make_simple_inventory()
        # Reading from disk
        self.assert_Getitems([b'dir1-id'], inv, [b'dir1-id'])
        self.assertTrue(b'dir1-id' in inv._fileid_to_entry_cache)
        self.assertFalse(b'sub-file2-id' in inv._fileid_to_entry_cache)
        # From cache
        self.assert_Getitems([b'dir1-id'], inv, [b'dir1-id'])
        # Mixed
        self.assert_Getitems([b'dir1-id', b'sub-file2-id'], inv,
                             [b'dir1-id', b'sub-file2-id'])
        self.assertTrue(b'dir1-id' in inv._fileid_to_entry_cache)
        self.assertTrue(b'sub-file2-id' in inv._fileid_to_entry_cache)

    def test_single_file(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b'TREE_ROOT', b'top-id'], inv, [b'top-id'])

    def test_get_all_parents(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b'TREE_ROOT', b'dir1-id', b'sub-dir1-id',
                           b'subsub-file1-id',
                           ], inv, [b'subsub-file1-id'])

    def test_get_children(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b'TREE_ROOT', b'dir1-id', b'sub-dir1-id',
                           b'sub-file1-id', b'sub-file2-id', b'subsub-file1-id',
                           ], inv, [b'dir1-id'])

    def test_from_root(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b'TREE_ROOT', b'dir1-id', b'dir2-id', b'sub-dir1-id',
                           b'sub-file1-id', b'sub-file2-id', b'sub2-file1-id',
                           b'subsub-file1-id', b'top-id'], inv, [b'TREE_ROOT'])

    def test_top_level_file(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b'TREE_ROOT', b'top-id'], inv, [b'top-id'])

    def test_subsub_file(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b'TREE_ROOT', b'dir1-id', b'sub-dir1-id',
                           b'subsub-file1-id'], inv, [b'subsub-file1-id'])

    def test_sub_and_root(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b'TREE_ROOT', b'dir1-id', b'sub-dir1-id', b'top-id',
                           b'subsub-file1-id'], inv, [b'top-id', b'subsub-file1-id'])


class TestMutableInventoryFromTree(TestCaseWithTransport):

    def test_empty(self):
        repository = self.make_repository('.')
        tree = repository.revision_tree(revision.NULL_REVISION)
        inv = mutable_inventory_from_tree(tree)
        self.assertEqual(revision.NULL_REVISION, inv.revision_id)
        self.assertEqual(0, len(inv))

    def test_some_files(self):
        wt = self.make_branch_and_tree('.')
        self.build_tree(['a'])
        wt.add(['a'], ids=[b'thefileid'])
        revid = wt.commit("commit")
        tree = wt.branch.repository.revision_tree(revid)
        inv = mutable_inventory_from_tree(tree)
        self.assertEqual(revid, inv.revision_id)
        self.assertEqual(2, len(inv))
        self.assertEqual("a", inv.get_entry(b'thefileid').name)
        # The inventory should be mutable and independent of
        # the original tree
        self.assertFalse(tree.root_inventory.get_entry(
            b'thefileid').executable)
        inv.get_entry(b'thefileid').executable = True
        self.assertFalse(tree.root_inventory.get_entry(
            b'thefileid').executable)


class ErrorTests(TestCase):

    def test_duplicate_file_id(self):
        error = DuplicateFileId('a_file_id', 'foo')
        self.assertEqualDiff(
            'File id {a_file_id} already exists in inventory as foo',
            str(error))
